--------------------------------------------------------------------------------
--  CURRENCY RELATED FUNCTIONS
--[[
	'bar' = Bar Number
		The dynamic table index of a group's record in: FI_SVPC_DATA["Groups"]
		This is a user interface convention to provide the appearance of sequential bar numbers regardless of the actual group ID.
	
	'gid' = Group ID
		The numeric primary key of a group as found in its data "record" stored at FI_SVPC_DATA.Groups
	
	'bid' = Button ID
		The numeric primary key of a button as found in its data "record" stored at FI_SVPC_DATA.Buttons
	
	'cid' = Currency ID
		The numeric primary key of a "watched" currency as found in its data "record" stored at FI_SVPC_DATA.Currencies
	
	'sid' = Session ID
		The numeric primary key of a farming session as found in its data "record" stored at FI_SVPC_DATA.Sessions
]]--
--------------------------------------------------------------------------------


function FI_HasTokens()
	-- check if player has any trackable currencies available
	for n=1,GetCurrencyListSize() do
		--name, isHeader, isExpanded, isUnused, isWatched, count, icon, maximum, hasWeeklyLimit, currentWeeklyAmount
		local name, isHeader, isExpanded, isUnused, isWatched, count = GetCurrencyListInfo(n);
		if name and (not isHeader) then
			return true;
		end
	end
end

--------------------------------------------------------------------------------
--  CURRENCY FUNCTIONS
--------------------------------------------------------------------------------
function FI_Currency( event )
	if FI_SV_CONFIG.debug then print("[FI_Currency]  Called."); end
	-- VERBOSE DEBUG
	--if FI_SV_CONFIG.debug then print("|cFFFF0000Currencies table:"); table.dump(FI_SVPC_DATA.Currencies); end

	-- configuration check
	if (not FI_SVPC_CONFIG.currency) then return; end
	
	-- one-time startup check
	if (FI_CURRENCY_LOADED == false) then
		FI_DB.check("Currencies");
		FI_CURRENCY_LOADED = true;
	end
	
	FI_Currency_Update(event);
	FI_Currency_Vis();
	
	-- VERBOSE DEBUG
	--if FI_SV_CONFIG.debug then print("|cFFFF0000Currencies table AFTER startup:"); table.dump(FI_SVPC_DATA.Currencies); end
	if FI_SV_CONFIG.debug then print("[FI_Currency]  Finished."); end
end

function FI_Currency_OnShow()
	-- configuration check
	if (not FI_SVPC_CONFIG.currency) then return; end
	
	if _G["BackpackTokenFrame"]:IsVisible() then
		if FI_SV_CONFIG.debug then print("[FI_Currency_OnShow]  Called."); end
		
		FI_MOVING = true;
		
		FI_Currency_Update();
		FI_Currency_Vis();
		
		FI_MOVING = false;
	end
end

function FI_Currency_Click( self, click, down )
	-- configuration check
	if (not FI_SVPC_CONFIG.currency) then return; end
	
	if (click == "RightButton") then
		local f_name = self:GetName(); --BackpackTokenFrameToken1
		local f_id = tonumber(strsub(f_name, -1));
		local eb_name = "FI_Currency_"..f_id.."_Editbox";

		if FI_SV_CONFIG.debug then print("You CLICKED the "..click.." ("..tostring(down)..") on frame:  "..f_name); end
		
		-- hide existing editboxes
		for cid=1,3 do
			local eb_n = "FI_Currency_"..cid.."_Editbox";
			if _G[eb_n] then _G[eb_n]:Hide(); end
		end
		
		-- create editbox if it doesn't exist yet
		if not _G[eb_name] then
			local s = FI_SVPC_STYLE.anchor;
			
			local f = CreateFrame("EditBox", eb_name, _G["BackpackTokenFrame"], "FI_TPL_Editbox");
			f:SetSize(70,35);
			f:SetJustifyH("CENTER");
			f:SetPoint("TOP", _G["BackpackTokenFrameToken"..f_id], "BOTTOM", 0, 0);
			
			f:SetScript("OnEnterPressed", FI_Currency_SetObjective);
			f:SetScript("OnEscapePressed", f.Hide);
			f:SetScript("OnEditFocusLost", f.Hide);
			
			f:SetMaxBytes(5);
			f:SetNumeric(true);
			f:SetAutoFocus(true);
			f:SetHistoryLines(6);
			
			f.text = f:CreateFontString("label","ARTWORK");
			f.text:SetPoint("LEFT", f, "LEFT", 12, 0);
			f.text:SetFont(s.text.font, s.text.size, s.text.flags);
			f.text:SetVertexColor(s.text.color[1], s.text.color[2], s.text.color[3]);
			f.text:SetAlpha(s.text.alpha);
			--f.text:SetText("Objective:");
		end
		
		-- get the current info
		local currency = FI_DB.select(FI_SVPC_DATA.Currencies, {id = f_id}, true);
		
		-- update the interface
		if currency then
			_G[eb_name]:SetNumber(currency.objective);
			_G[eb_name]:Show();
			_G[eb_name]:HighlightText();
				
			--FI_Message("Enter a |cFF"..FI_SV_CONFIG.Colors.currency[4]..currency.name.."|r objective at the bottom of your backpack.");
		elseif FI_SV_CONFIG.debug then
			print("[FI_Currency_Click]  DB query failed for 'f_id':  "..f_id); --debug
		end
	end
end

--------------------------------------------------------------------------------

function FI_Currency_Update( event )
	if FI_SV_CONFIG.debug then print("[FI_Currency_Update]  Called."); end
	
	-- take a snapshot of current data
	local temp = table.copy(FI_SVPC_DATA.Currencies);
	
	-- process all three backpack "watched currency" slots
	for cid=1,3 do
		-- API query
		local name,count,icon = GetBackpackCurrencyInfo(cid);
		
		local results;
		if name then
			-- server returned something for the current slot
			local query = {
				["name"] = name,
				["count"] = count,
			}
			
			-- check for lastcount data in snapshot
			local existing = FI_DB.select(temp, {["name"] = name}, true);
			if existing then
				query["lastcount"] = existing.count;
			else
				query["lastcount"] = count;
			end

			-- check for cached objective data
			if FI_SVPC_CACHE.Currencies then
				local cached = FI_DB.select(FI_SVPC_CACHE.Currencies, {["name"] = name}, true);
				if cached then
					query["objective"] = cached.objective;
				else
					query["objective"] = 0;
					query["success"] = false;
				end
			end
			
			------------------------------------------------------------
			-- save updated record
			------------------------------------------------------------
			results = FI_DB.update(FI_SVPC_DATA.Currencies, {id = cid}, query);
		
		elseif (FI_MOVING == true) and (FI_LOADING == false) then
			------------------------------------------------------------
			-- clear unused "slot"
			------------------------------------------------------------
			results = FI_Currency_Clear(cid);
			
		else
			------------------------------------------------------------
			-- use existing data
			------------------------------------------------------------
			results = FI_DB.select(FI_SVPC_DATA.Currencies, {id = cid}, true);
		end
		
		------------------------------------------------------------
		-- update interface
		------------------------------------------------------------
		local f_name = "FI_Currency_"..cid;
		_G[f_name.."_Objective"]:SetText(results.objective);
		
		------------------------------------------------------------
		-- progress tracking
		------------------------------------------------------------
		if (FI_LOADING == false) then
			FI_Progress(results);
		end
	end
end

function FI_Currency_Vis()
	if FI_SV_CONFIG.debug then print("[FI_Currency_Vis]  Called."); end
	
	for cid=1,3 do
		local currency = FI_DB.select(FI_SVPC_DATA.Currencies, {id = cid}, true);
		
		if currency then
			local f_name = "FI_Currency_"..cid;
			
			--_G[f_name.."_Objective"]:SetText(currency.objective); --shouldnt normally be necessary to do this here...
			
			if currency.name and (currency.objective > 0) then
				-- set color
				local color;
				if (currency.count < currency.objective) then
					color = FI_SV_CONFIG.Colors.objective;
				else
					color = FI_SV_CONFIG.Colors.success;
				end
				
				_G[f_name.."_Objective"]:SetVertexColor(color[1], color[2], color[3]);
				
				-- set visibility
				_G[f_name]:Show();
			
			else
				-- set color
				color = FI_SV_CONFIG.Colors.objective;
				_G[f_name.."_Objective"]:SetVertexColor(color[1], color[2], color[3]);
				
				-- set visibility
				_G[f_name]:Hide();

				if FI_SV_CONFIG.debug then print("[FI_Currency_Vis]  Hiding frame "..cid); end
			end
			
			
		elseif FI_SV_CONFIG.debug then
			print("[FI_Currency_Vis]  ERROR! Unable to find currency record for slot "..cid); --debug
		end
	end
end

-- FI_SetObjective wrapper
function FI_Currency_SetObjective( editbox, id, input )
	local cid,amt;
	
	if editbox then
		cid = FI_FrameToID( editbox:GetName() );
		amt = editbox:GetNumber();
		
	elseif id and input then
		cid = tonumber(id);
		amt = tonumber(input);
		
	elseif FI_SV_CONFIG.debug then
		-- error
		print("[FI_Currency_SetObjective]  Input error."); --debug
	end
	
	-- select record
	local currency = FI_DB.select(FI_SVPC_DATA.Currencies, {["id"] = cid}, true);
	
	if currency then
		-- skip updating if input matches current objective
		if (amt ~= currency.objective) then
			-- dew it!
			FI_SetObjective("Currencies", currency.id, amt);
			FI_Currency_Update();
		
		elseif FI_SV_CONFIG.debug then
			print("[FI_Currency_SetObjective]  Duplicate objective input, update skipped."); --debug
		end
			
	elseif FI_SV_CONFIG.debug then
		print("[FI_Currency_SetObjective]  DB query failed for CID:  "..cid); --debug
	end
	
	--hide input frame
	editbox:Hide();
end

-- reset a currency record
function FI_Currency_Clear( cid )
	-- reset data
	local query = {
		["name"] = "",
		["count"] = 0,
		["lastcount"] = 0,
		["objective"] = 0,
		["success"] = false,
	}
	
	local results = FI_DB.update(FI_SVPC_DATA.Currencies, {id = cid}, query);
	
	-- update interface
	local f_name = "FI_Currency_"..cid;
	_G[f_name.."_Objective"]:SetText(0);
	_G[f_name]:Hide();
	
	return results;
end

-- check database structure
function FI_Currency_Check()
	-- validate currency data
	if (GetNumWatchedTokens() > 0) then
		if FI_SV_CONFIG.debug then print("[FI_Currency_Check]  Validating currency records..."); end
		
		local result;
		
		for cid=1,3 do
			local name,count,icon = GetBackpackCurrencyInfo(cid);
			local currency = FI_DB.select(FI_SVPC_DATA.Currencies, {id = cid}, true);
			
			if name then
				if (name == currency.name) then
					result = true;
				else
					result = false; break;
				end
			
			elseif FI_SV_CONFIG.debug then
				print("[FI_Currency_Check]  API query failed on slot "..cid); --debug
				result = false; break;
			end
		end
		
		-- returns false if there is a problem
		return result;
	end
end


--------------------------------------------------------------------------------
--  MONEY TRACKING
--------------------------------------------------------------------------------
function FI_Money( event, arg )
	if FI_SV_CONFIG.debug then FI_Message("FI_Money:  "..event.."  "..arg); end
end
